use super::api::ErrorModel;
use std::fmt::{Display, Formatter};
use std::io;

#[derive(Debug)]
pub enum GenericError {
    TomlError(toml::de::Error),
    YepzonServerError(ErrorModel),
    MinreqError(minreq::Error),
    ChronoParseError(chrono::ParseError),
    IOError(io::Error),
    FromUTF8Error(std::string::FromUtf8Error),
    GPXError(gpx::errors::Error),
    MessagedError(String, Option<Box<GenericError>>),
}

impl From<toml::de::Error> for GenericError {
    fn from(error: toml::de::Error) -> Self {
        GenericError::TomlError(error)
    }
}

impl From<minreq::Error> for GenericError {
    fn from(error: minreq::Error) -> Self {
        GenericError::MinreqError(error)
    }
}

impl From<chrono::ParseError> for GenericError {
    fn from(error: chrono::ParseError) -> Self {
        GenericError::ChronoParseError(error)
    }
}

impl From<ErrorModel> for GenericError {
    fn from(error: ErrorModel) -> Self {
        GenericError::YepzonServerError(error)
    }
}

impl From<io::Error> for GenericError {
    fn from(error: io::Error) -> Self {
        GenericError::IOError(error)
    }
}

impl From<std::string::FromUtf8Error> for GenericError {
    fn from(error: std::string::FromUtf8Error) -> Self {
        GenericError::FromUTF8Error(error)
    }
}

impl From<gpx::errors::Error> for GenericError {
    fn from(error: gpx::errors::Error) -> Self {
        GenericError::GPXError(error)
    }
}

impl Display for GenericError {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), std::fmt::Error> {
        let err = match self {
            GenericError::ChronoParseError(e) => format!("Date-Time value parse error: {}", e),
            GenericError::FromUTF8Error(e) => format!("UTF-8 error: {}", e),
            GenericError::IOError(e) => format!("IO error: {}", e),
            GenericError::MinreqError(e) => format!("Network error: {}", e),
            GenericError::TomlError(e) => format!("Toml error: {}", e),
            GenericError::YepzonServerError(e) => format!(
                "Yepzon server error: {}",
                e.message.as_ref().unwrap_or(&String::new())
            ),
            GenericError::MessagedError(msg, err) => format!(
                "{}\n  {}",
                msg,
                err.as_ref().map(|e| e.to_string()).unwrap_or(String::new())
            ),
            GenericError::GPXError(e) => format!("GPX error: {}", e),
        };
        write!(f, "{}", err)
    }
}

pub trait MessagedError<A> {
    fn with_msg<T: Into<String>>(self, text: T) -> Result<A, GenericError>;
}

impl<A, B: Into<GenericError>> MessagedError<A> for Result<A, B> {
    fn with_msg<T: Into<String>>(self, text: T) -> Result<A, GenericError> {
        match self {
            Ok(ok) => Ok(ok),
            Err(e) => Err(GenericError::MessagedError(
                text.into(),
                Some(Box::new(e.into())),
            )),
        }
    }
}
